/*
 * SPDX-FileCopyrightText: Copyright (c) 2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
 * SPDX-License-Identifier: BSD-2-Clause
 */

// Templated addition function: myadd
template <typename T>
__device__ T myadd(T a, T b) { return a + b; }
